# todo
import os

import streamlit as st
from auto_app import APIAutoApp
st.title("FastAPI Microservices Management")

# Configuration file selection
st.header("Configuration")
config_files = st.multiselect("Select configuration files", os.listdir("."))
environment = st.selectbox("Select environment", ["development", "production"])

api_auto_app = APIAutoApp(config_files)

# Start/stop microservices
st.header("Microservices Control")
start_button = st.button("Start Microservices")
stop_button = st.button("Stop Microservices")

if start_button:
    api_auto_app.run()

if stop_button:
    # Implement logic to stop microservices
    pass

# Additional functionality
st.header("Additional Functionality")
# Add your custom Streamlit components or interactable elements here




import logging
from typing import TypeVar, Any, Optional, List, Callable, Dict, Tuple

from fastapi import APIRouter, status
from pydantic import BaseModel

logger = logging.getLogger(__name__)
T = TypeVar('T')


# Custom Exceptions
class RouteException(Exception):
    pass


class UnsupportedMethod(RouteException):
    pass


class PathConflict(RouteException):
    pass


class AutoRouteManager:
    _instance = None
    registered_paths = set()

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.registered_paths = set()
        return cls._instance


def get_default_path_and_method(function: Callable, methods: str, keyword_mapping: Dict[str, str]) -> Tuple[str, str]:
    """
    Generate a default path for a given function and mapped keywords.

    The function name is split on underscores, and each part is checked against the mapped keywords. If a match is found,
    the part is replaced with a placeholder using the keyword.

    Args:
        function: The function for which the path is to be generated.
        methods: The default HTTP method if not provided explicitly.
        keyword_mapping: A dictionary of keywords to placeholders.

    Returns:
        The generated path as a string and the extracted method as a string.
    """

    HTTP_METHODS_AVAILABLE = ['get', 'post', 'put', 'delete', 'options', 'head', 'patch', 'trace']
    logger.debug(f'{function.__name__ = }')
    path_parts = function.__name__.split('_')

    method_place_holder = path_parts[0].lower() in ['get', 'post', 'put', 'delete', 'options', 'head', 'patch',
                                                    'trace']

    if methods:
        default_method = methods[0].lower()
        if method_place_holder and path_parts[0].lower() != default_method:
            logger.warning(f"The function name suggests the HTTP method {path_parts[0].upper()}, "
                           f"but the passed method is {methods[0].upper()}. "
                           f"Consider changing the function name to match the method for clarity.")
    else:
        assert len(path_parts) > 0, "Function name must be at least one character long."
        assert path_parts[0].lower() in HTTP_METHODS_AVAILABLE, \
            f"Function name must start with a valid HTTP method. " \
            f"Valid methods are: {HTTP_METHODS_AVAILABLE}"
        default_method = path_parts[0].lower()

    path_parts = path_parts if not method_place_holder else path_parts[1:]
    for i, part in enumerate(path_parts):
        if part in keyword_mapping:
            path_parts[i] = "{" + keyword_mapping[part] + "}"

    generated_default_path = f"/{default_method}/{'/'.join(path_parts)}"

    logger.debug(f'generated_default_path: {generated_default_path}')
    logger.debug(f'default_method: {default_method}')
    logger.debug(f'path_parts: {path_parts}')
    logger.debug(f'methods: {methods}')

    # exit()
    return generated_default_path, default_method


def check_path(path: str, override: bool) -> None:
    registered_paths = AutoRouteManager()._instance.registered_paths
    if not override and path in registered_paths:
        raise PathConflict(f"Path {path} is already registered. To override, set `override=True`."
                           f"The current Registered Paths are: {registered_paths}")
    registered_paths.add(path)


def route_functions_in_class(cls, router: APIRouter, prefix: str, tags: List[str]) -> None:
    for attr_name, attr_value in cls.__dict__.items():
        if callable(attr_value) and hasattr(attr_value, '_auto_route'):
            path, kwargs = attr_value._auto_route
            for method in kwargs.pop('methods'):
                route_decorator = getattr(router, method.lower(), None)
                if route_decorator is None:
                    raise UnsupportedMethod(f"Unsupported HTTP method: {method}")
                if tags:
                    kwargs['tags'] = kwargs.get('tags', []) + tags
                if prefix:
                    print(f'prefix: {prefix}')
                    path = f"{prefix.rstrip('/')}{path}"
                route_decorator(path, **kwargs)(attr_value)


def auto_route(
        path: Optional[str] = None,
        response_model: Optional[TypeVar] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        status_code: Optional[int] = status.HTTP_200_OK,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        override: Optional[bool] = False,
        mapped_keywords: Optional[dict] = None,
        **kwargs: Any,
) -> Callable[..., T]:
    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        nonlocal path, response_model, description, tags, status_code, operation_id, override, mapped_keywords, summary

        mapped_keywords = mapped_keywords or {}
        path, method = get_default_path_and_method(func, kwargs.pop('methods', None), mapped_keywords)
        check_path(path, override)

        func._auto_route = (path, {
            'response_model': response_model or func.__annotations__.get('return'),
            'description': description or func.__doc__,
            'tags': tags or [],
            'status_code': status_code,
            'operation_id': operation_id or func.__name__,
            'summary': summary or func.__name__.replace('_', ' ').title(),
            'methods': [method],
            **kwargs,
        })

        # return staticmethod(func)
        return func

    return decorator


def register_router(router: APIRouter, prefix: str = '', tags: List[str] = []):
    def decorator(cls):
        nonlocal tags
        if not tags:
            tags = [cls.__name__.replace('_', ' ')]
        route_functions_in_class(cls, router, prefix, tags)
        return cls

    return decorator


class APIAutoRouter:
    def __init__(self):
        self.router = APIRouter()

    def route(self, *args, **kwargs):
        return auto_route(*args, **kwargs)

    def register(self, prefix: str = '', tags: List[str] = []):
        return register_router(self.router, prefix, tags)

    @staticmethod
    def retrieve_from_env(api_key_name):
        # if not available in environment variables, use the .env file
        import os
        if not os.environ.get(api_key_name):
            from dotenv import load_dotenv
            load_dotenv()
        return os.environ.get(api_key_name)


if __name__ == '__main__':
    api = APIAutoRouter()


    class User(BaseModel):
        id: int
        name: str


    @api.register()
    class User_Controller:
        @api.route()
        def get_user_by_id(id: int):
            return {"id": id, "name": "User"}

        @api.route()
        def post_user_by_id(user: User):
            return {"id": user.id, "name": user.name}

        @api.route()
        def delete_user_by_id(id: int):
            return {"id_deleted": id}


    from auto_app import APIAutoApp

    APIAutoApp(routers_list=[api.router]).run(host='localhost', port=8000)



import logging
import os

from auto_app import APIAutoApp


def main():
    import argparse
    parser = argparse.ArgumentParser(description='Run FastAPI application.')
    parser.add_argument('--config', type=str, required=False, help='Path to the configuration file.')
    parser.add_argument('--env_type', type=str, default="development", required=False,
                        help='Environment to run the application in.')
    parser.add_argument('--host', type=str, default="0.0.0.0", help='Host to run the application on.')
    parser.add_argument('--port', type=int, default=8000, help='Port to run the application on.')
    parser.add_argument('--run', action='store_true', help='Run the application.')
    parser.add_argument('--demo', action='store_true',
                        help='Run the Streamlit demo. (alpha - does not accept any of the other kwargs for running the app')
    args = parser.parse_args()

    if args.demo:  # todo make this use the microservice rather than running with its own settings, UI should just be UI
        import subprocess
        subprocess.run(["streamlit", "run", "route_generation/streamlit_demo.py"])
        return
    if not os.path.isfile(args.config):
        logging.error(f"Configuration file {args.config} does not exist.")
        raise FileNotFoundError(f"Configuration file {args.config} does not exist.")

    apiautoapp = APIAutoApp()

    if args.run:
        app = apiautoapp.build_app()
        apiautoapp.run(host=args.host, port=args.port, app=app)
    else:
        return apiautoapp.build_app()


if __name__ == "__main__":
    main()



def create_composed_class(name, classes, strategy='namespaced'):
    """
        Dynamically create a class composed of methods from other classes.

        This function creates a new class composed of methods from other classes.
        In the case of a method name collision, the resolution strategy depends on
        the strategy argument.

        Args:
            name (str): The name of the composed class to be created.
            classes (tuple): A tuple of classes from which to compose the new class.
            strategy (str, optional): The strategy for resolving method name collisions.
                'namespaced' - Prefix each method with its class name.
                'prioritize' - Use the method from the class specified first in the classes tuple.
                Defaults to 'namespaced'.

        Returns:
            type: The composed class.

        Raises:
            ValueError: If an invalid strategy is specified.

        Example:
            >>> class A:
            >>>     def method(self):
            >>>         return "Method from A"
            >>>
            >>> class B:
            >>>     def method(self):
            >>>         return "Method from B"
            >>>
            >>> # Create the composed class with namespaced methods
            >>> C = create_composed_class("C", (A, B), strategy='namespaced')
            >>> c = C()
            >>> print(c.A_method())  # prints: "Method from A"
            >>> print(c.B_method())  # prints: "Method from B"
            >>>
            >>> # Create the composed class with prioritized methods
            >>> D = create_composed_class("D", (A, B), strategy='prioritize')
            >>> d = D()
            >>> print(d.method())  # prints: "Method from A"
        """
    methods = {}
    for cls in classes:
        for method_name, method in cls.__dict__.items():
            if callable(method) and not method_name.startswith("__"):
                if strategy == 'namespaced':
                    method_name = f"{cls.__name__}_{method_name}"
                elif strategy == 'prioritize':
                    if method_name in methods:
                        continue

                def method_wrapper(self, *args, **kwargs):
                    instance = getattr(self, f"_{cls.__name__}")
                    return method(instance, *args, **kwargs)

                methods[method_name] = method_wrapper

    ComposedClass = type(name, (object,), methods)

    def __init__(self):
        for cls in classes:
            setattr(self, f"_{cls.__name__}", cls())

    setattr(ComposedClass, "__init__", __init__)

    return ComposedClass




# sudo /home/ruben/anaconda3/envs/AIOrg/bin/python setup.py install --user
 # curl -s "http://localhost:8000/openapi.json" > openapi.json && docker run --rm -i -v ${PWD}:/local openapitools/openapi-generator-cli generate -i local/openapi.json -g python -o /local/pythondemoapi --additional-properties packageName=pythondemoapi && rm -f openapi.json


from setuptools import setup, find_packages

setup(
    name='auto_route',
    version='0.1.6',  # Semantic Versioning
    packages=["auto_route", "auto_app"],
    author='Ruben Fernandez',
    author_email='ruben@carbonyl.org',
    description="Thank you for your interest in contributing to the AutoRoute project! Our goal is to simplify and "
                "automate the creation of SaaS/PaaS solutions. We're excited to welcome you to our community.",
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    url='https://github.com/Bucanero06/auto_route',
    install_requires=[
        "fastapi",
        "pydantic",
        "uvicorn",
    ],
)



import importlib
import logging
import os
import sys
from typing import List, Optional

from fastapi import FastAPI, APIRouter
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel



class InitClassObject(BaseModel):
    class_name: str
    init_args: Optional[dict] = None


class AutoLoadRouterConfigObject(BaseModel):
    module_path: str
    router_names: List[str]

    # Classes to instantiate
    init_classes: Optional[List[InitClassObject]] = None


class APIAutoApp:
    def __init__(self, routers_list: List[APIRouter and str] = None):
        self.routers_list = routers_list

    def load_routers_to_app(self, app: FastAPI, routers) -> None:

        for router in routers:
            from auto_route.auto_route import APIAutoRouter
            if isinstance(router, AutoLoadRouterConfigObject):
                # fixme relative paths
                module_path = router.module_path.replace(".py", "").replace("/", ".")
                current_dir = os.path.dirname(__file__)
                parent_dir = os.path.abspath(os.path.join(current_dir, os.pardir))
                # path_to_add = os.path.abspath(os.path.join(current_dir, module_path))
                sys.path.insert(0, parent_dir)

                module = importlib.import_module(module_path, package=".")
                for router_name in router.router_names:
                    custom_api_router = getattr(module, router_name)
                    app.include_router(router=custom_api_router.router)
                #
                # Init classes if any
                if router.init_classes is not None:  # This is so fucking stupid to-do
                    assert isinstance(router.init_classes[0], InitClassObject), "init_class_names must be a list"
                    for iclass in router.init_classes:
                        service_class = getattr(module, iclass.class_name)
                        service_instance = service_class(**(iclass.init_args or {}))
                        logging.debug(f"Initialized {iclass.class_name} as {service_instance}")
            #
            #
            elif isinstance(router, APIRouter or APIAutoRouter):
                app.include_router(router=router)
            else:
                print(
                    f'Router input {router} is not a valid type expected for routes. Either AutoLoadRouterConfigObject or APIRouter')

    def build_app(self, settings: dict = None) -> FastAPI:
        if settings is None:
            settings = {}
        app = FastAPI(**settings)

        app.add_middleware(
            CORSMiddleware,
            allow_origins=settings.get('allow_origins', ["*"]),
            allow_credentials=settings.get('allow_credentials', True),
            allow_methods=settings.get('allow_methods', ["*"]),
            allow_headers=settings.get('allow_headers', ["*"]),
        )

        self.load_routers_to_app(app, self.routers_list)
        return app

    def run(self, host: str = "0.0.0.0", port: int = 8000, app: FastAPI = None, **kwargs) -> None:
        if app is None:
            app = self.build_app()
        import uvicorn
        uvicorn.run(app, host=host, port=port, **kwargs)


if __name__ == '__main__':
    stripe_router_config = AutoLoadRouterConfigObject(
        module_path='Stripe/stripe_class.py',
        router_names=['stripe_router'],
        init_classes=[
            InitClassObject(class_name='Customer'),
            InitClassObject(class_name='Payment_Intent'),
            InitClassObject(class_name='Subscriptions'),
        ]
    )

    apiautoapp = APIAutoApp(routers_list=[stripe_router_config])

    app = apiautoapp.build_app()
    apiautoapp.run(host='localhost', port=8000, app=app)
